import 'dart:async';
import 'package:adeptannotations/adeptannotations.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:source_gen/source_gen.dart';

class CloneGenerator
    extends GeneratorForAnnotation<Cloneable> {

  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement = _getClassElement(element);
    String cloneFunction = _generateCloneFunction(classElement, annotation);
    return '''
    $cloneFunction
    ''';
  }

  final _annotationChecker = const TypeChecker.fromRuntime(CloneKey);

  _checkDefaultConstructor(ClassElement classElement) {
    bool hasDefaultConstructor = false;
    classElement.constructors?.forEach((constructor) {
      if (constructor?.isDefaultConstructor ?? false) {
        hasDefaultConstructor = true;
        return;
      }
    });
    if (!hasDefaultConstructor)
      throw InvalidGenerationSourceError('To use @Cloneable make sure '
          'to have a default constructor in your class '
          '${classElement.displayName}.',
        todo: 'Add a datult constructor like MyClass() or '
                'MyClass({String param1, String param2}) to your '
                '${classElement.displayName} or remove the @Cloneable '
                'annotation from it.',
        element: classElement);
  }

  ClassElement _getClassElement(Element element) {
    if (element is! ClassElement)
      throw InvalidGenerationSourceError('Generator cannot target '
        '${element.displayName}. @Cloneable can only be applied to a '
          'class.',
        todo: 'Check the type of ${element.displayName} or remove the '
          'annotation from it.',
        element: element);
    return element as ClassElement;
  }

  String _generateCloneFunction(
      ClassElement classElement, ConstantReader annotation) {
    _checkDefaultConstructor(classElement);
    String classType = classElement?.displayName;
    String functionName = "\$${classType}Clone";
    String currentDataParam = "currentData";
    String newDataParam = "newData";
    String mergeParam = "merge";
    String functionParams = "{$classType $currentDataParam, "
        "$classType $newDataParam, bool $mergeParam = false}";
    String cloneVar = "clone";

    StringBuffer cloneData = StringBuffer("final $cloneVar = $classType()");
    StringBuffer mergeData = StringBuffer("if($mergeParam ?? false){");
    classElement?.fields?.forEach((field) {
      CloneKey cloneKey = _getCloneKey(field);
      String fieldName = field?.displayName;
      String cloneDataFieldValue = "$newDataParam?.$fieldName ?? "
          "${currentDataParam}.$fieldName";
      if(cloneKey?.forceNewData ?? false)
        cloneDataFieldValue = "$newDataParam?.$fieldName";
      cloneData.writeln("..${fieldName} = $cloneDataFieldValue");
      mergeData.writeln("${currentDataParam}.$fieldName = $cloneVar?.$fieldName;");
    });
    cloneData.writeln(";");
    mergeData.writeln("}");

    return '''
      /// This function creates a new instance of [$classType] by creating a 
      /// copy of [$currentDataParam] and replacing its fields with the data in 
      /// [$newDataParam].
      /// Notes:
      /// 1. The function will use the values from [$newDataParam], in case of 
      /// null then $currentDataParam values will be used. 
      /// 2. If [$mergeParam] is true then [$currentDataParam] will mirror the new 
      /// instance returned.
      /// 3. If @CloneKey(forceNewData: true) annotation is used in some field, 
      /// then [$newDataParam] value will be forced for the specified field 
      /// no matter it's nullability 
      $classType $functionName($functionParams){
        ${cloneData.toString()}
        ${mergeData.toString()}
        return $cloneVar;
      }
    ''';
  }

  CloneKey _getCloneKey(FieldElement field) {
    int annotationsCount = _annotationChecker
        .annotationsOfExact(field)
        ?.length ?? 0;
    if(annotationsCount > 1)
      throw InvalidGenerationSourceError(
        'Repeated `@CloneKey` annotation on ${field?.displayName}.',
        todo: 'Check the `@CloneKey` annotations on field '
          '${field?.displayName}.',
        element: field);
    if (annotationsCount == 1)
      return CloneKey(
        forceNewData: _annotationChecker?.firstAnnotationOfExact(field)
          ?.getField('forceNewData')
          ?.toBoolValue()
      );
    return null;
  }
}
